package com.google.code.fastbuilder.internal.bytecode;

import static org.objectweb.asm.Opcodes.*;

import org.objectweb.asm.*;

import com.google.code.fastbuilder.internal.*;

public class BytecodeBuilderClassVisitor extends ClassAdapter {

	private static final String CONSTRUCTOR_NAME = "<init>";
	private static final String VOID_METHOD = "()V";

	private final BuilderSpecification builderSpecification;
	private final String builderInterfaceName;
	private final String builderImplementationName;

	private String targetInternalName;
	private String targetDescriptor;

	public BytecodeBuilderClassVisitor(final ClassVisitor visitor, final Class<?> target, final BuilderSpecification builderSpecification, final String builderInterfaceName,
			final String builderImplementationName) {
		super(visitor);

		this.builderSpecification = builderSpecification;
		this.builderInterfaceName = builderInterfaceName;
		this.builderImplementationName = builderImplementationName;
		this.targetInternalName = Type.getInternalName(target);
		this.targetDescriptor = Type.getDescriptor(target);
	}

	@Override
	public void visit(final int version, final int access, final String name, final String signature, final String superName, final String[] interfaces) {
		super.visit(version, ACC_PUBLIC + ACC_FINAL + ACC_SUPER, builderImplementationName, null, superName, new String[] { builderInterfaceName });
		generateConstructor();
	}

	@Override
	public MethodVisitor visitMethod(final int access, final String name, final String desc, final String signature, final String[] exceptions) {
		if (builderSpecification.isWriter(name, (Type.getArgumentsAndReturnSizes(desc) >> 2) - 1)) {
			Type fieldType = Type.getArgumentTypes(desc)[0];
			generateField(fieldType.getDescriptor(), name);
			generateWriter(fieldType, name, desc, signature, exceptions);
		} else {
			Type returnType = Type.getReturnType(desc);
			generateReader(returnType, name, desc, signature, exceptions);
		}
		return null;
	}

	@Override
	public void visitEnd() {
		generateBuild();
		super.visitEnd();
	}

	protected void generateConstructor() {
		MethodVisitor mv = super.visitMethod(ACC_PUBLIC, CONSTRUCTOR_NAME, VOID_METHOD, null, null);
		mv.visitCode();
		mv.visitVarInsn(ALOAD, 0);
		mv.visitMethodInsn(INVOKESPECIAL, Type.getInternalName(Object.class), CONSTRUCTOR_NAME, VOID_METHOD);

		mv.visitInsn(RETURN);
		mv.visitMaxs(1, 1);
		mv.visitEnd();
	}

	protected void generateField(final String type, final String name) {
		generateField(ACC_PRIVATE, type, name);
	}

	protected void generateField(final int access, final String type, final String name) {
		FieldVisitor fv = super.visitField(access, name, type, null, null);
		fv.visitEnd();
	}

	protected void generateWriter(final Type type, final String name, final String desc, final String signature, final String[] exceptions) {
		MethodVisitor mv = super.visitMethod(ACC_PUBLIC, name, desc, signature, exceptions);
		mv.visitCode();
		mv.visitVarInsn(ALOAD, 0);
		mv.visitVarInsn(type.getOpcode(ILOAD), 1);
		mv.visitFieldInsn(PUTFIELD, builderImplementationName, name, type.getDescriptor());
		mv.visitVarInsn(ALOAD, 0);
		mv.visitInsn(ARETURN);
		mv.visitMaxs(2, 2);
		mv.visitEnd();
	}

	protected void generateReader(final Type type, final String name, final String desc, final String signature, final String[] exceptions) {
		MethodVisitor mv = super.visitMethod(ACC_PUBLIC, name, desc, signature, exceptions);
		mv.visitCode();
		mv.visitVarInsn(ALOAD, 0);
		mv.visitFieldInsn(GETFIELD, builderImplementationName, builderSpecification.getFieldNameFromReader(name), type.getDescriptor());
		mv.visitInsn(type.getOpcode(IRETURN));
		mv.visitMaxs(1, 1);
		mv.visitEnd();
	}

	protected void generateBuild() {
		MethodVisitor mv = super.visitMethod(ACC_PUBLIC, "build", "()" + targetDescriptor, null, null);
		mv.visitCode();
		mv.visitTypeInsn(NEW, targetInternalName);
		mv.visitInsn(DUP);
		mv.visitVarInsn(ALOAD, 0);
		mv.visitMethodInsn(INVOKESPECIAL, targetInternalName, CONSTRUCTOR_NAME, "(L" + builderInterfaceName + ";)V");
		mv.visitInsn(ARETURN);
		mv.visitMaxs(3, 1);
		mv.visitEnd();

		mv = super.visitMethod(ACC_PUBLIC + ACC_BRIDGE + ACC_SYNTHETIC, "build", "()" + Type.getDescriptor(Object.class), null, null);
		mv.visitCode();
		mv.visitVarInsn(ALOAD, 0);
		mv.visitMethodInsn(INVOKEVIRTUAL, builderImplementationName, "build", "()L" + targetInternalName + ";");
		mv.visitInsn(ARETURN);
		mv.visitMaxs(1, 1);
		mv.visitEnd();
	}
}
