package com.google.code.fastbuilder.internal.bytecode;

import java.io.*;
import java.util.*;
import java.util.concurrent.*;

import org.objectweb.asm.*;
import org.objectweb.asm.util.*;
import org.slf4j.*;

import com.google.code.fastbuilder.*;
import com.google.code.fastbuilder.internal.*;
import com.google.code.fastbuilder.internal.classloading.*;
import com.google.code.fastbuilder.util.*;

public class BytecodeBuilderProvider implements BuilderProvider {

	private static final Logger logger = LoggerFactory.getLogger(BytecodeBuilderProvider.class);

	private static final String BUILDER_SUFFIX = "Generated";

	private final Map<Class<?>, Class<? extends IBuilder<?>>> builderClassCache;

	public BytecodeBuilderProvider() {
		builderClassCache = new ConcurrentHashMap<Class<?>, Class<? extends IBuilder<?>>>();
	}

	@Override
	@SuppressWarnings("unchecked")
	public <T, V extends IBuilder<T>> V make(final Class<V> spec, final BuilderSpecification builderSpec) {
		Class<? extends IBuilder<?>> builderClass = builderClassCache.get(spec);
		if (builderClass == null) { // It's ok to regenerate builders if concurrent access occurs
									// because they are exactly the same
			builderClass = (Class<? extends IBuilder<?>>) generateBuilderClass(spec, builderSpec);
			builderClassCache.put(spec, builderClass);
		}
		try {
			return (V) builderClass.newInstance();
		} catch (Exception e) {
			// Should not happen
			throw new RuntimeException(e);
		}
	}

	@SuppressWarnings("unchecked")
	protected <T, V extends IBuilder<T>> Class<V> generateBuilderClass(final Class<T> spec, final BuilderSpecification builderSpec) {
		Class<?> target = Types.getTypeArgumentOfInterface(spec, IBuilder.class);
		String builderInterfaceName = Type.getInternalName(spec);
		String builderImplementationName = builderInterfaceName + BUILDER_SUFFIX;
		BuilderClassLoader classLoader = ClassLoaders.getClassLoader(target);

		logger.debug("Generating builder implementation class [{}] for target [{}]", builderImplementationName, target);

		ClassReader reader;
		ClassWriter writer;
		try {
			reader = new ClassReader(builderInterfaceName);
			writer = new ClassWriter(0);
			ClassVisitor chainVisitor = writer;
			if (logger.isDebugEnabled()) {
				TraceClassVisitor traceClassVisitor = new TraceClassVisitor(writer, new PrintWriter(System.out));
				chainVisitor = traceClassVisitor;
			}

			BytecodeBuilderClassVisitor generator = new BytecodeBuilderClassVisitor(chainVisitor, target, builderSpec, builderInterfaceName, builderImplementationName);
			reader.accept(generator, ClassReader.SKIP_DEBUG);
		} catch (IOException e) {
			throw new IllegalStateException("Unable to generate builder bytecode", e);
		}
		return (Class<V>) classLoader.defineClass(builderImplementationName.replace('/', '.'), writer.toByteArray());
	}
}
